##################################################
#####     IMPORTS                            #####
##################################################

#####  EXTERNAL IMPORTS
# ENV AND PATHS
import os
import shutil
import platform

# RUNNING COMMANDS
import sys
import subprocess

# CLI
import argparse
from yaspin import yaspin

##### INTERNAL IMPORTS
# ERRORS
from .__errors__ import *

# APP
from .app import App
from .set_up import SetUp
##################################################
#####     CODE                               #####
##################################################

from .__vars__ import settings_paths

#####  DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

##################################################
#####     CLASS                              #####
##################################################

## CUSTOM ARGPARSE ACTIONS
# Update action
class UpdateAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        self.__update()
        print("The program has been updated successfully")
        exit()

    # Update if app is a python module
    @yaspin(text="Updating application...")
    def __update(self) -> None:
        if "WorkSpaceAutomation" in sys.modules:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'install', '-U', 'WorkSpaceAutomation'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT
            )
        else:
            raise NotPythonModule("The program is not installed as a python module, update must be done by manually.")

# Uninstall action
class UninstallAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        self.__uninstall()
        print("The program has been uninstalled successfully")
        exit()

    # Delete all files/folders and uninstall if its a python module
    @yaspin(text="Uninstalling application...")
    def __uninstall(self) -> None:
        shutil.rmtree(MAIN_DIRECTORY, ignore_errors=True)
        if "WorkSpaceAutomation" in sys.modules:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'uninstall', 'WorkSpaceAutomation'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT
            )
        else:
            raise NotPythonModule("The program is not installed as a python module, uninstall must be done by manually. (Program data was already errased)")

## PARSER
class Parser:

    def __init__(self) -> None:

        ## PARSER
        self.parser = argparse.ArgumentParser(
            usage=f"""
%(prog)s [-h] [-v]

Set-Up:
%(prog)s [-i] [-k KEY] [--vscode VS_CODE] [--set-up-languages LANGUAGES] [--custom-dir CUSTOM_DIR] [--create-folders] [--github-user GITHUB_USER] [--api-key API_KEY] [-y]

Command:
%(prog)s [-c COMMAND] [-k KEY] [--name NAME] [--directory DIRECTORY] [--language LANGUAGE] [--github GITHUB] [--owner OWNER] [--private PRIVATE] [--license LICENSE] [--new-directory NEW_DIRECTORY] [--add-apps ADD_APPS] [--del-apps DEL_APPS] [--add-urls ADD_URLS] [--del-urls DEL_URLS] [-y]

SubCommand:
%(prog)s [-c config] [-sc SUB_COMMAND] [-k KEY] [--vscode VS_CODE] [--setup-languages LANGUAGES] [--custom-dir CUSTOM_DIR] [--create-folders] [--github-user GITHUB_USER] [--api-key API_KEY] [-y]""",
            formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=90)
        )

        ## GENERAL ARGUMENTS
        self.parser.add_argument('-v', '--version', action='version', version='Current version is ' + "0.1.0", help='Gives the version of the program')
        self.parser.add_argument('--update', action=UpdateAction, help='Update the program to latest released version (Only if the program is a python module)')
        self.parser.add_argument('--uninstall', action=UninstallAction, help='Uninstall the program, this action will not errase your workspaces')

        ## MAIN
        self.parser.add_argument('-i', '--init', action='store_true', help='Initialise the program')
        self.parser.add_argument('-c', '--command', action='store', type=str, help='Enter a command to run')
        self.parser.add_argument('-sc', '--sub-command', action='store', type=str, help='Enter a sub-command if required')
        self.parser.add_argument('-k', '--key', action='store', type=str, help='Enter your key')
        self.parser.add_argument('-y', '--yes', action='store_true', help='Confirm')

        ## SET-UP
        self.parser.add_argument('--vscode', action='store', type=str, help='Enter your VSCode version type (code/insiders)')
        self.parser.add_argument('--set-up-languages', action='store', type=str, help='Enter a comma separated list of languages to install')
        self.parser.add_argument('--custom-dir', action='store', type=str, help='Path if you want to use a custom directory')
        self.parser.add_argument('--create-directories', action='store_true', help='Set if you want to create new folders for the workspaces')
        self.parser.add_argument('--github-user', action='store', type=str, help='Enter your GitHub username')
        self.parser.add_argument('--api-key', action='store', type=str, help='Enter your GitHub API key')

        ### APP
        ## APP INFO
        self.parser.add_argument('--name', action='store', type=str, help='Enter your project name')
        self.parser.add_argument('--directory', action='store', type=str, help='Enter the directory of your project')
        self.parser.add_argument('--language', action='store', type=str, help='Enter the language of your project')

        ## APP GITHUB
        self.parser.add_argument('--github', action='store_true', help='Enter if you want to create a GitHub repo')
        self.parser.add_argument('--owner', action='store', type=str, help='Enter the owner of the repo')
        self.parser.add_argument('--private', action='store_true', help='Set if you want to create a private repo')
        self.parser.add_argument('--license', action='store', type=str, help='Enter the license of the repo')

        ## APP MOVE
        self.parser.add_argument('--new-directory', action='store', type=str, help='Enter the directory you want to move the workspace to')

        ## APP EDIT
        self.parser.add_argument('--add-apps', action='store_true', help='Command for adding new apps to the workspace')
        self.parser.add_argument('--del-apps', action='store_true', help='Command for deleting apps from the workspace')
        self.parser.add_argument('--add-urls', action='store_true', help='Command for adding new URLs to the workspace')
        self.parser.add_argument('--del-urls', action='store_true', help='Command for deleting URLs from the workspace')

    # Main Function
    def main(self) -> None:

        arguments = vars(self.parser.parse_args())

        # Check if key is provided
        if not arguments.get("key"):
            raise PasswordNotProvided("The password must be provided.")

        # Check if app was already initialised
        if not os.path.exists(MAIN_DIRECTORY) and not arguments.get("init"):
            raise AppNotInitialised("""The app is not set-up. Please run "--init" command.""")

        # Run the program with the provided arguments
        if arguments.get("init"):
            SetUp(
                key=arguments["key"],
                kwargs=arguments
            )

        elif arguments.get("command"):
            App(
                key=arguments["key"],
                command=arguments["command"],
                sub_command=arguments["sub_command"],
                kwargs=arguments
            )

# Entry point for the CLI program
def entry_point() -> None:

    # Creates the parser and runs it
    cli = Parser()
    cli.main()

#####  RUN FILE
if __name__ == "__main__":
    entry_point()



##################################################
#####     IMPORTS                            #####
##################################################

#####  EXTERNAL IMPORTS
# READING APP DATA
from dotenv import load_dotenv
from cryptography.fernet import Fernet
import json

# ENV MANAGEMENT
import os
import platform

##### INTERNAL IMPORTS
# ERRORS
from .__errors__ import *

# CLI
from .Config import ConfigFunctions
from .WorkspaceManager import WorkspaceFunctions

# REUSABLE FUNCTIONS
from .common import key_generator

##################################################
#####     CODE                               #####
##################################################

#####  IMPORT GLOBAL VARIABLES FROM FILE
from .__vars__ import settings_paths

#####  DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

#####  DEFINE MAIN FILES DIRECTORIES
SETTINGS = os.path.join(MAIN_DIRECTORY, "settings.json")
SECRETS = os.path.join(MAIN_DIRECTORY, ".secrets")

##################################################
#####     CLASS                              #####
##################################################

class App:

    def __init__(self,
        key: str,
        command: str,
        sub_command: str = None,
        **kwargs
        ) -> None:

        # Set-up App variables
        self.yes = kwargs.get("yes") or False
        self.key = key
        self.__load_env_variables()
        self.__load_settings()

        # Accepted commands
        commands = [
            "create",
            "import",
            "delete",
            "open",
            "publish",
            "move",
            "edit",
            "config"
        ]

        # Sub Commands for config command
        sub_commands = [
            "vscode",
            "languages",
            "directories",
            "api-key",
            "git-user"
        ]

        # Check if the command is a non-config command
        if command in commands.remove("config"):
            WorkspaceFunctions(
                command=command,
                api_key=self.API_KEY,
                settings=self.settings,
                yes=self.yes,
                kwargs=kwargs
            )

        # Check for a sub-command for the config command
        elif command == "config" and sub_command in sub_commands:

            config = ConfigFunctions(
                command=command,
                sub_command=sub_commands,
                settings=self.settings,
                yes=self.yes,
                kwargs=kwargs
            )

            # Save the change in a dictionary
            change = {sub_command: config.parameter}

            # Save changed settings
            ConfigFunctions().save_settings(
                settings=self.settings,
                kwargs=change
            )

        # The command given was incomplete or incorrect
        else:
            raise CommandNotFound("The given command was not found or not given. Use --help for more info.")

    # Get saved API_KEY
    def __load_env_variables(self) -> None:

        # Try to load the API_KEY
        try:
            load_dotenv(SECRETS)
            encrypted_key = os.getenv("API_KEY")
            self.API_KEY = Fernet(key_generator(self.key)).decrypt(encrypted_key).decode()

        # Raise FileNotFoundError if the .secrets file does not exist
        except FileNotFoundError:
            raise FileNotFoundError(".secrets file was not found")

        # Raise PasswordError if the password is incorrect
        else:
            raise PasswordError("Password is incorrect, API KEY could not get decrypted. Check if it was correctly typed.")

    # Get saved settings
    def __load_settings(self) -> None:

        # Raise an error if the Settings file does not exist
        if not os.path.exists(SETTINGS):
            raise FileNotFoundError("Settings file was not found.")

        with open(SETTINGS, "r+") as f:
            self.settings = json.load(f)



##################################################
#####     IMPORTS                            #####
##################################################

#####  EXTERNAL IMPORTS
# ENV AND PATHS
import os
import platform

##### INTERNAL IMPORTS
# ERRORS
from .__errors__ import *

# CLI
from .Config import ConfigFunctions

##################################################
#####     CODE                               #####
##################################################

from .__vars__ import settings_paths

#####  DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

##################################################
#####     CLASS                              #####
##################################################

class SetUp:

    def __init__(self,
        key: str,
        **kwargs
        ) -> None:

        # OS of the user
        self.__check_compatibility()

        # Check if a setup has already been performed
        if os.path.exists(MAIN_DIRECTORY):
            raise ReSetupError("Cannot re-setup, maybe you rather run '-c config -sc <sub_command>' command.")
        else:
            os.makedirs(MAIN_DIRECTORY, exist_ok=True)

        # Create the config values
        config = ConfigFunctions(
            key=key,
            command="init",
            kwargs=kwargs
        )

        # Variables related to settings.json
        self.directories = config.directories
        self.languages = config.languages
        self.vscode = config.vscode
        self.git_username = config.git_user

        # GitHub api key
        self.api_key = config.api_key

        # Save settings to their respective locations
        ConfigFunctions().save_settings(kwargs={
            "directories": self.directories,
            "languages": self.languages,
            "vscode": self.vscode,
            "github-user": self.git_username,
            "api-key": self.api_key
        })

    # Determine if os is compatible
    def __check_compatibility(self) -> None:

        # Compatible OS
        if platform.system() == "Linux":
            return None
        elif platform.system() == "Darwin":
            return None
        elif platform.system() == "Windows":
            return None

        # Incompatibility Error
        else:
            raise NotCompatibleOS("Your OS is not supported.")



########################################
#####  IMPORTING MODULES           #####
########################################

#####  EXTERNAL IMPORTS
import base64
import hashlib

from time import sleep
import os

########################################
#####  CODE                        #####
########################################

#####  FUNCTIONS
# Clear the screen
def clear(t: float):
    os.system('cls' if os.name == 'nt' else 'clear')
    sleep(t)

# Generate a hashlib key for the Fernet class
def key_generator(key: str):
    return base64.urlsafe_b64encode(hashlib.md5(key.encode()).hexdigest().encode())



##################################################
#####     ERRORS                             #####
##################################################

## GENERAL ERRORS
class NotPythonModule(Exception): ...

## SET-UP ERRORS
class AppNotInitialised(Exception): ...
class ReSetupError(Exception): ...
class NotCompatibleOS(Exception): ...

## APP ERRORS
# Main Errors
class PasswordNotProvided(Exception): ...
class CommandNotFound(Exception): ...
class PasswordError(Exception): ...
class MissingArguments(Exception): ...

# WorkSpace Errors
class WorkSpaceAlreadyExists(Exception): ...
class WorkSpaceNotFound(Exception): ...

# Config Errors
class DirectoryNotFound(Exception): ...
class API_KEY_ERROR(Exception): ...
class InvalidSubCommand(Exception): ...



########################################
#####  IMPORTING MODULES           #####
########################################

#####  EXTERNAL IMPORTS
# ENV MANAGEMENT
import os
import platform

# HTTP REQUESTS
import requests

# CLI
from yaspin import yaspin

##### INTERNAL IMPORTS
# ERRORS
from ...__errors__ import *

########################################
#####  GLOBAL VARIABLES            #####
########################################

##### IMPORT GLOBAL VARIABLES FROM FILE
from ...__vars__ import settings_paths

GITHUB_API = "https://api.github.com" # GitHub API URL

##### DEFINE MAIN DIRECTORY
MAIN_DIRECTORY = settings_paths[platform.system()]

########################################
#####  CLASS                       #####
########################################

class Github:

    def __init__(self, action: int, clone: bool, API_KEY: str, name:str, owner: str, directory: str, **kwargs) -> None:

        self.TOKEN = API_KEY

        self.headers = {
        "Authorization": "Bearer " +  self.TOKEN,
        "Accept": "application/vnd.github+json",
        "X-GitHub-Api-Version": "2022-11-28"
        }

        functions = [
            self.__create_user_repo,
            self.__create_org_repo,
            self.__delete_repo
        ]

        if kwargs:
            try:
                private = kwargs.get("private")
                auto_init = kwargs.get("auto_init")
                gitignore = kwargs.get("gitignore") if auto_init == "true" else None
                license = kwargs.get("license") if auto_init == "true" else None
            except:
                raise MissingArguments("There are missing arguments, please re-run the command")

        self.directory = directory
        self.owner = owner
        self.name = name

        if action in (0,1):
            if auto_init == "true":
                payload = '{"name":"' + self.name + '","private":' + private + ',"auto_init":' + auto_init + ',"gitignore_template":"' + gitignore +  '","license_template":"' + license + '"}'
            else:
                payload = '{"name":"' + self.name + '","private":' + private + ',"auto_init":' + auto_init + '"}'
            functions[action](payload)

        elif action == 2:
            functions[action]()

        if clone:
            self.__clone_repo()

    @yaspin(text="Creating GitHub Repository...")
    def __create_user_repo(self, payload) -> None:
        requests.post(GITHUB_API+"/user/repos", data=payload, headers=self.headers)

    @yaspin(text="Creating GitHub Repository...")
    def __create_org_repo(self, payload) -> None:
        requests.post(GITHUB_API+f"/orgs/{self.owner}/repos", data=payload, headers=self.headers)

    @yaspin(text="Deleting GitHub Repository...")
    def __delete_repo(self) -> None:
        requests.delete(GITHUB_API+f"/repos/{self.owner}/{self.name}", headers=self.headers)

    @yaspin(text="Cloning GitHub Repository...")
    def __clone_repo(self) -> None:
        os.chdir(os.path.join(self.directory, self.name))
        url = f"https://github.com/{self.owner}/{self.name}.git"
        os.system("git clone" + url)
        os.chdir(MAIN_DIRECTORY)

    @yaspin(text="Pushing project to github")
    def push_user_repo(self) -> None:
        os.chdir(os.path.join(self.directory, self.name))

        if not os.path.exists("README.md"):
            os.system('ECHO #' + self.name + ' >> README.md')

        os.system("git init")
        os.system("git add . && git commit -m 'Initial Commit' && git branch -M main")
        os.system(f"git remote add origin http://{self.TOKEN}@github.com/{self.owner}/{self.name}.git")
        os.system("git push -u origin main")
        os.chdir(MAIN_DIRECTORY)



########################################
#####  IMPORTING MODULES           #####
########################################

#####  EXTERNAL IMPORTS
# ENV MANAGEMENT
import shutil
import os
import platform

# CLI
from yaspin import yaspin

########################################
#####  CODE                        #####
########################################

#####  GLOBAL VARIABLES
from ...__vars__ import settings_paths, languages

##### DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

########################################
#####  CLASS                       #####
########################################

class ContentsManager:

    ##### INITIALISE CLASS
    @yaspin(text="Setting up WorkSpace...")
    def __init__(self,
        action: int,
        directory: str,
        name: str,
        language: str = None,
        new_directory: str = None
        ) -> None:

        # All actions in a list
        actions = [
            self.create_workspace,
            self.move_workspace,
            self.delete_workspace
        ]

        # Set given parameters as class variables
        self.directory = directory
        self.name = name
        self.language = language
        self.new_directory = new_directory

        # Create new class variables from given parameters
        self.workspace_folder = os.path.join(self.directory, self.name)
        self.new_workspace_folder = os.path.join(self.new_directory, self.name) or None

        # Execute action requested
        actions[action]()

    ##### CREATE ALL DIRECTORIES AND FILES ACCORDING TO THE LANGUAGE SELECTED
    def create_workspace(self) -> None:

        os.makedirs(self.workspace_folder, exist_ok=True)
        os.chdir(self.workspace_folder)

        directories: list = languages[self.language]["directories"]
        files: list = languages[self.language]["files"]
        venv_command: str = languages[self.language]["venv-command"]

        directories.extend([
            ".vscode"
        ])

        files.extend([
            ".vscode/settings.json"
        ])

        for directory in directories:
            os.makedirs(directory)

        for file in files:
            with open(file, "x+") as f:
                pass

        if venv_command:
            os.system(venv_command)

    ##### MOVE SELECTED WORKSPACE TO NEW LOCATION
    def move_workspace(self) -> None:
        shutil.move(self.workspace_folder, self.new_workspace_folder)
        self.delete_workspace()

    ##### DELETE ALL DIRECTORIES AND FILES FROM A WORKSPACE
    def delete_workspace(self) -> None:
        shutil.rmtree(self.workspace_folder)



########################################
#####  IMPORTING MODULES           #####
########################################

#####  EXTERNAL IMPORTS
import subprocess
import os
import platform

########################################
#####  GLOBAL VARIABLES            #####
########################################

##### IMPORT GLOBAL VARIABLES FROM FILE
from ...__vars__ import settings_paths

##### DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

##### DEFINE MAIN FILES DIRECTORIES
WORKSPACES = os.path.join(MAIN_DIRECTORY, "workspaces.json")

########################################
#####  FUNCTION                    #####
########################################

def run_apps(directory: str, apps: list, profile: str) -> None:

    # Run VSCode with selected profile on selected directory
    subprocess.run(["powershell.exe", "-c", f"code-insiders --profile {profile} {directory}"])

    # Try to run GitHub on selected directory
    try:
        subprocess.run(["powershell.exe", "-c", f"github {directory}"])
    except:
        pass

    # Run all apps for selected WorkSpace
    for app in apps:

        # Try to run the app
        try:
            subprocess.run(app)
        except:
            pass



########################################
#####  GLOBAL VARIABLES            #####
########################################

__version__ = "0.1.0" # Current Version of the project

# Long description
__description__ = "A project to help all fellow devs to not waste time in doing the boring and repetitive stuff"

# Other Metadata
__author__ = "Dtar380"
__license__ = "GPL v3.0"
__status__ = "4 - Beta"



########################################
#####  IMPORTING MODULES           #####
########################################

##### EXTERNAL IMPORTS
# READING APP DATA
import json

# ENV MANAGEMENT
import subprocess
import os
import platform

# URL REQUESTS
import requests

# CLI
import inquirer
from yaspin import yaspin

##### INTERNAL IMPORTS
# ERRORS
from ..__errors__ import *

########################################
#####  CLASS                       #####
########################################

class SharedMenus:

    # GET INFO
    # Get info to create a WorkSpace
    @classmethod
    def get_workspace_info(cls, directories: list, languages: list, get_language: bool, **kwargs) -> tuple:

        name = kwargs.get("name") or inquirer.text("Enter the name of your WorkSpace")
        directory = kwargs.get("directory") or inquirer.list_input("Select the directory for your WorkSpace", choices=directories)
        github = kwargs.get("github") or inquirer.confirm("Do you want to create a GitHub repository?", default=True)

        if get_language:
            language = kwargs.get("language") or inquirer.list_input("Select the main language for the project", choices=languages)
        else:
            language = None

        return name, directory, github, language

    # Get info to create GitHub repo
    @classmethod
    def get_github_info(cls, default_owner: str, get_license: bool, **kwargs) -> tuple:

        owner = kwargs.get("owner") or inquirer.text("Enter the name of the owner of the repository", default=default_owner)
        private = kwargs.get("private") or "true" if inquirer.confirm("Do you want to create a private repository?", default=False) else "false"

        if get_license:
            license = kwargs.get("license") if kwargs.get("license") in ["MIT", "GPL-3.0", "Unlicense"] else inquirer.list_input("Enter a license", choices=["MIT", "GPL-3.0", "Unlicense"])
        else:
            license = None

        return owner, private, license

    # Get ingo to move a WorkSpace to another Path
    @classmethod
    def get_move_workspace_info(cls, directories: list, **kwargs) -> tuple:

        name = kwargs.get("name") or inquirer.text("Enter the name of your WorkSpace")
        new_directory = kwargs.get("new_directory") if kwargs.get("new_directory") in directories else inquirer.list_input("Select the new directory for the WorkSpace", choices=directories)

        return name, new_directory

    # SELECTION MENUS
    # Select APPs to boot when opening a WorkSpace
    @classmethod
    def select_apps(cls) -> dict:

        command_generators = {
            "Windows": cls.__find_app_windows,
            "Darwin": cls.__find_app_macos,
            "Linux": cls.__find_app_linux
        }

        apps = {}

        while True:

            app = inquirer.text(message="Enter the name of the APP")

            command = command_generators[platform.system()](app)

            if not command:
                print(f"The App '{app}' Was not found")
            elif inquirer.confirm(message=f"Want to use the APP {app}"):
                apps[app] = command

            if not inquirer.confirm(message="Want to add more APPs?"):
                return apps

    # Find APPs in each supported OS
    # Find app pn Windows
    @yaspin(text=" Finding Application...")
    def __find_app_windows(self, app_name: str) -> list:

        try:
            file = subprocess.run(["powershell.exe", "-c", 'Get-StartApps | Where-Object {{ $_.Name -like "{}" }}'.format(app_name)], capture_output=True)
            app = file.stdout.decode().replace("\r\n","").split(" ")[-1]

            if app:
                return ["powershell.exe", "-c", 'Start-Process Shell:AppsFolder\\{}'.format(app)]
            else:
                return None

        except:
            return None

    # Find app on macOS
    @yaspin(text=" Finding Application...")
    def __find_app_macos(self, app_name: str):

        try:
            result = subprocess.run(['ls', '/Applications'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            apps = [line for line in result.stdout.splitlines() if app_name.lower() in line.lower()]

            if apps:
                return ['open', f"/Applications/{apps[0]}"]
            else:
                return None

        except:
            return None

    # Find app on Linux
    @yaspin(text=" Finding Application...")
    def __find_app_linux(self, app_name: str):

        try:
            result = subprocess.run(['which', app_name], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            app_path = result.stdout.strip()

            if app_path:
                return [app_path]
            else:
                return None

        except:
            return None

    # Select URLs to open when opening a WorkSpace
    @classmethod
    def select_urls(cls) -> list:

        urls = []

        while True:
            url = inquirer.text("Enter a URL")

            if requests.get(url).status_code != 200:
                print(f"{url} was unable to fetch")
            elif inquirer.confirm(message=f"Want to add the URL {url}"):
                urls.append(url)

            if not inquirer.confirm(message="Want to add more URLs"):
                return urls

    # GET OPENING INFO
    # Get VSCode profile to open with the WorkSpace
    @classmethod
    def select_vscode_profile(cls, vscode_settings: str) -> str:

        with open(vscode_settings, "r+") as f:
            profiles = json.load(f)["userDataProfiles"]

        profile_names = [profile["name"] for profile in profiles]

        profiles = inquirer.Checkbox(
            "profiles",
            message="Selecciona el perfil de VSCode:",
            choices=profile_names
        )

        return inquirer.prompt([profiles])["profiles"]



########################################
#####  IMPORTING MODULES           #####
########################################

##### EXTERNAL IMPORTS
# READING APP DATA
import json

# ENV MANAGEMENT
import os
import platform

# CLI
import inquirer
from yaspin import yaspin

##### INTERNAL IMPORTS
# ERRORS
from ..__errors__ import *

# CLI
from .shared_menus import SharedMenus

# BUILDERS
from .builders import Github
from .builders import ContentsManager

# RUNNERS
from .runners import run_apps
from .runners import open_urls

########################################
#####  GLOBAL VARIABLES            #####
########################################

##### IMPORT GLOBAL VARIABLES FROM FILE
from ..__vars__ import settings_paths, languages

##### DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

##### DEFINE MAIN FILES DIRECTORIES
WORKSPACES = os.path.join(MAIN_DIRECTORY, "workspaces.json")

########################################
#####  CLASS                       #####
########################################

class WorkspaceFunctions:

    ##### INITIALISE CLASS
    def __init__(self,
        command: str,
        api_key: str,
        settings: dict,
        yes: bool = False,
        **kwargs
        ) -> None:

        self.api_key = api_key
        self.settings = settings
        self.yes = yes

        # Valid commands
        commands = {
            "create": self.create_workspace,
            "import": self.import_workspace,
            "delete": self.delete_workspace,
            "open": self.open_workspace,
            "publish": self.publish_workspace,
            "move": self.move_workspace,
            "edit": self.edit_workspace,
        }

        # Run selected command
        commands[command](**kwargs)

    ##### WORKSPACE FUNCTIONS
    # Create a new WorkSpace
    def create_workspace(self, **kwargs) -> None:

        name, directory, github, language = SharedMenus.get_workspace_info(
            directories=self.settings["directories"],
            languages=self.settings["languages"].keys(),
            get_language=True,
            kwargs=kwargs
        )

        if name in os.listdir(directory):
            raise WorkSpaceAlreadyExists("A WorkSpace with this name already exists.")

        if github:
            owner, private, license = SharedMenus.get_github_info(
                default_owner=self.settings["github_user"],
                get_license=True,
                kwargs=kwargs
            )

            self.__create_github_repo(
                name=name,
                owner=owner,
                directory=directory,
                private=private,
                auto_init="true",
                clone=False,
                license=license,
                language=language
            )

        ContentsManager(
            action=1,
            directory=directory,
            name=name,
            language=language
        )

        apps = SharedMenus.select_apps()
        urls = SharedMenus.select_urls()

        self.__save_workspace(
            directory=directory,
            name=name,
            apps=apps,
            urls=urls,
            owner=owner or None,
            private=private or None
        )

    # Import an exiting project as a WorkSpace
    def import_workspace(self, **kwargs) -> None:

        name, directory, github, language = SharedMenus.get_workspace_info(
            directories=self.settings["directories"],
            languages=self.settings["languages"].keys(),
            get_language=False,
            kwargs=kwargs
        )

        if name in os.listdir(directory):
            raise WorkSpaceAlreadyExists("A WorkSpace with this name already exists.")

        if github:
            owner, private, license = SharedMenus.get_github_info(
                default_owner=self.settings["github_user"],
                get_license=False,
                kwargs=kwargs
            )

            self.__create_github_repo(
                name=name,
                owner=owner,
                directory=directory,
                private=private,
                auto_init="false",
                clone=False
            )

        apps = SharedMenus.select_apps()
        urls = SharedMenus.select_urls()

        self.__save_workspace(
            directory=directory,
            name=name,
            apps=apps,
            urls=urls,
            owner=owner or None,
            private=private or None
        )

    # Delete an existing WorkSpace
    def delete_workspace(self, **kwargs) -> None:

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        name = kwargs.get("name") or inquirer.text("Enter the name of your WorkSpace")
        if name not in data:
            raise WorkSpaceNotFound("There was no WorkSpace found with that name.")

        directory = data[name]["directory"]
        owner = data[name]["owner"]

        if inquirer.confirm(f"Want to delete the WorkSpace {name}?") and self.yes:
            ContentsManager(
                actions=2,
                directory=directory,
                name=name
            )

        elif inquirer.confirm(f"Want to delete the repo {name}?") and self.yes:
            Github(
                action=2,
                clone=False,
                API_KEY=self.api_key,
                name=name,
                owner=owner,
                directory=directory
            )

        else:
            print("exiting...")
            return None

        del data[name]

        with open(WORKSPACES, "w+") as f:
            json.dump(data, f)

    # Open VSCode, APPs and URLs associated to a WorkSpace
    def open_workspace(self, **kwargs) -> None:

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        name = kwargs.get("name") or inquirer.text("Enter the name of your WorkSpace")
        if name not in data.keys():
            raise WorkSpaceNotFound("There was no WorkSpace found with that name.")

        directory = data[name]["directory"]
        profile = SharedMenus.select_vscode_profile(vscode_settings=self.settings["vscode"]["path"])

        apps = data[name]["apps"].values()
        urls = data[name]["urls"]

        # RUNNERS
        run_apps(directory=directory, apps=apps, profile=profile)
        open_urls(urls=urls)

    # Publish a WorkSpace to GitHub
    def publish_workspace(self, **kwargs) -> None:

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        name = kwargs.get("name") or inquirer.text("Enter the name of your WorkSpace")
        owner, private, license = SharedMenus.get_github_info(
            default_owner=self.settings["github_user"],
            get_license=True,
            kwargs=kwargs
        )

        if name not in data.keys():
            raise WorkSpaceNotFound("There was no WorkSpace found with that name.")

        language = data[name]["language"] or None

        self.__create_github_repo(
            name=name,
            owner=owner,
            directory=data[name]["directory"],
            private=private,
            auto_init="false",
            clone=False,
            license=license,
            language=language
        )

        data[name]["owner"] = owner
        data[name]["private"] = private

        with open(WORKSPACES, "w+") as f:
            json.dump(data, f)

    # Move an existing WorkSpace to another Path
    def move_workspace(self, **kwargs) -> None:

        name, new_directory = SharedMenus.get_move_workspace_info(
            directories=self.settings["directories"],
            kwargs=kwargs
        )

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        if name not in data.keys():
            raise WorkSpaceNotFound("There was no WorkSpace found with that name.")

        if not inquirer.confirm(f"Want to move the WorkSpace {name}?") and not self.yes:
            print("Exiting...")
            return None

        ContentsManager(
            action=1,
            directory=data[name]["directory"],
            name=name,
            new_directory=new_directory
        )

        data[name]["directory"] = os.path.join(new_directory, name)

        with open(WORKSPACES, "w+") as f:
            json.dump(data, f)

    # Edit the information of an existing repo
    def edit_workspace(self, **kwargs) -> None:

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        # ADD APPS TO WORKSPACE
        if kwargs.get("add-apps"):
            apps = SharedMenus.select_apps()
            data["apps"] += apps

        # REMOVE APPS FROM WORKSPACE
        if kwargs.get("del-apps"):
            apps_to_delete = inquirer.Checkbox("apps", message="Please select the APPs to remove.", choices=data["apps"].keys())
            for app in inquirer.prompt([apps_to_delete])["apps"]:
                del data["apps"][app]

        # ADD URLS TO WORKSPACE
        if kwargs.get("add-urls"):
            urls = SharedMenus.select_urls()
            data["urls"].extend(urls)

        # DELETE URLS FROM WORKSPACE
        if kwargs.get("del-urls"):
            urls_to_delete = inquirer.Checkbox("urls", message="Please select the URLs to remove.", choices=data["urls"])
            for url in inquirer.prompt([urls_to_delete])["urls"]:
                data["urls"].remove(url)

        if not inquirer.confirm("Want to perform changes?") and not self.yes:
            print("Exiting...")
            return None

        with open(WORKSPACES, "r+") as f:
            data = json.dump(data, f)

    # Save a WorkSpace thats being currently editted
    def __save_workspace(self,
        directory: str,
        name: str,
        apps: dict,
        urls: list,
        owner: str = None,
        private: str = None
        ) -> None:

        if not inquirer.confirm("Want to create the WorkSpace?") and not self.yes:
            print("Exiting...")
            return None

        spinner = yaspin(text=" Saving the WorkSpace...")
        spinner.start()

        with open(WORKSPACES, "r+") as f:
            data = json.load(f)

        data[name] = {
            "directory": os.path.join(directory, name),
            "apps": apps,
            "urls": urls,
            "owner": owner,
            "private": private
        }

        with open(WORKSPACES, "w+") as f:
            json.dump(data, f)

        spinner.stop()

    # Create a GitHub repo
    def __create_github_repo(self,
        name: str,
        owner: str,
        directory: str,
        private: str,
        auto_init: str,
        clone: bool,
        license: str = None,
        language: str = None,
        ) -> None:

        if not inquirer.confirm("Want to create the repo?") and not self.yes:
            print("Exiting...")
            return None

        gitignore = languages[language] or None

        action = 0 if owner == self.settings["git-user"] else 1

        github = Github(
            action=action,
            clone=clone,
            API_KEY=self.api_key,
            name=name,
            owner=owner,
            directory=directory,
            private=private,
            auto_init=auto_init,
            license=license,
            gitignore=gitignore
        )

        github.push_user_repo()



########################################
#####  IMPORTING MODULES           #####
########################################

##### EXTERNAL IMPORTS
# READING APP DATA
from pathlib import Path
from cryptography.fernet import Fernet
import json
import os

# ENV MANAGEMENT
import platform

# HTTP REQUESTS
import requests

# CLI LIBRARIES
import inquirer
from yaspin import yaspin

##### INTERNAL IMPORTS
# ERRORS
from ..__errors__ import *

# COMMON
from ..common import key_generator, clear

########################################
#####  GLOBAL VARIABLES            #####
########################################

##### IMPORT GLOBAL VARIABLES FROM FILE
from ..__vars__ import code_paths, settings_paths, languages

##### DEFINE MAIN DIRECTORY ACCORDING TO OPERATING SYSTEM
MAIN_DIRECTORY = settings_paths[platform.system()]

##### DEFINE MAIN FILES DIRECTORIES
WORKSPACES = os.path.join(MAIN_DIRECTORY, "workspaces.json")
SETTINGS = os.path.join(MAIN_DIRECTORY, "settings.json")
SECRETS = os.path.join(MAIN_DIRECTORY, ".secrets")

########################################
#####  CLASS                       #####
########################################

class ConfigFunctions:

    ##### INITIALISE CLASS
    def __init__(self,
        key: str,
        command: str,
        sub_command: str = None,
        settings: dict = None,
        yes: bool = False,
        **kwargs) -> None:

        # Set-up the class variables
        self.command = command
        self.key = key
        self.settings = settings
        self.yes = yes

        # Accepted commands
        commands = {
            "vscode": self.vscode_type,
            "languages": self.select_languages,
            "directories": self.select_directories,
            "github-user": self.get_github_username,
            "api-key": self.get_api_key
        }

        # If the main command is init run every sub-command
        if self.command == "init":
            self.directories = commands["directories"](kwargs)
            clear(0.5)
            self.languages = commands["languages"](kwargs)
            clear(0.5)
            self.vscode = commands["vscode"](kwargs)
            clear(0.5)
            self.git_user = commands["github-user"](kwargs)
            clear(0.5)
            self.api_key = commands["api-key"](kwargs)
            clear(0.5)

        # If there was a valid subcommand provided is config run the sub-command
        elif sub_command in commands.keys():
            self.parameter = commands[sub_command](kwargs)

        # Else, raise an error
        else:
            raise InvalidSubCommand("The sub-command is not valid or not provided.")

    ##### VSCODE
    # Select VSCode type between insiders and normal
    def vscode_type(self, **kwargs) -> dict:

        # Get the VSCode type and convert it into directory and PATH ENV VARIABLE
        code_type = kwargs.get("vscode") or inquirer.list_input("Select the VSCode type", choices=["code", "insiders"])

        code_dir = "code - insiders" if code_type == "insiders" else "code"
        code_PATH = "code-insiders" if code_type == "insiders" else "code"

        vscode = {
            "type": code_PATH,
            "path": code_paths[platform.system()].format(code_dir)
        }

        # Ask for confirmation
        if not inquirer.confirm("Want to save changes?") and not self.yes and self.command != "init":
                print("Exiting...")
                return None

        # Return the VSCode dict with the directory and PATH ENV VARIABLE
        return vscode

    ##### LANGUAGES
    # Select the most used languages by the user
    def select_languages(self, **kwargs) -> list:

        # Get the languages and convert it into a list
        all_languages = languages.keys()

        # Get the preselected languages
        selected_languages = self.settings["languages"] or kwargs.get("set_up_languages") or None

        # Ask the user to select the languages using a checkbox
        question = inquirer.Checkbox(
            name = "languages",
            message = "Select the languages you use",
            choices = all_languages,
            default = selected_languages)

        answer = inquirer.prompt([question])["languages"]

        # Ask for confirmation
        if not inquirer.confirm("Want to save changes?") and not self.yes and self.command != "init":
            print("Exiting...")
            return None

        # Return the selected languages
        return answer

    ##### DIRECTORIES RELATED
    # Get the directories to work on with the workspaces
    def select_directories(self, **kwargs) -> tuple:

        # Get the main directory and if user wants to use existing directories or create new ones
        main_dir = kwargs.get("custom_dir") or Path(Path.joinpath(Path.home(), "Documents"))
        create = kwargs.get("create_directories") or inquirer.confirm("Want to create new directories?", default=False)

        # Raise an error if the main directory was not found
        if not os.path.exists(main_dir):
            raise DirectoryNotFound("Specified directory was not found.")

        # Get the directories in the main directory
        directories = [i.lower() for i in os.listdir(main_dir) if "." not in i]

        # If there are existing directories in main directory and user doesn't want to create new ones
        if directories and not create:
            sub_directories = self.__select_directories(directories=directories)
        else:
            sub_directories = self.__ask_directories()
            self.__create_directories(
                main_dir=main_dir,
                sub_directories=sub_directories)

        # Return main directory and sub directories selected as a tuple
        return main_dir, sub_directories

    # Get user selected directories
    def __select_directories(self, directories) -> list:

        # Set github directory as default if exists
        if "github" in directories:
            default = ["github"]

        # Ask the user to select the directories using a checkbox
        question = inquirer.Checkbox(
            name = "directories",
            message = "Select the directories you want to use",
            choices = directories,
            default = default)

        answer = inquirer.prompt([question])["directories"]

        # Return the selected directories as a list
        return answer

    # Get directories that user wants to create
    def __ask_directories(self, main_directory: str) -> list:

        sub_directories = []

        # Get directories until user wants to stop
        while True:
            folder = inquirer.text(message="Enter the name for the folder")

            # Check if the directory already exists and ask for confirmation
            if os.path.exists(os.path.join(main_directory, folder)):
                print(f"Folder {folder} already exists, enter a new name.")
            elif inquirer.confirm(message=f"Want to create a folder called {folder}?"):
                sub_directories.append(folder)

            if not inquirer.confirm(message="Want to add more directories?"):
                return sub_directories

    # Create the new directories
    @yaspin(text=" Creating directories...")
    def __create_directories(self, main_dir: str, sub_directories: list) -> None:

        # Ask for confirmation
        if not inquirer.confirm("Want to save changes?") and not self.yes and self.command != "init":
            print("Exiting...")
            return None

        # Create directories
        for i in sub_directories:
            os.mkdir(os.path.join(main_dir, i))

    ##### API KEY
    # Get Api Key and return encrypted version
    def get_api_key(self, **kwargs) -> str:

        # Get API Key
        new_key = kwargs.get("api_key") or inquirer.text("Enter your API Key")

        # Verify API Key
        headers = {
            "Authorization": "Bearer " + new_key,
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28"
        }

        message = requests.get(url="https://api.github.com/repos/Dtar380/WorkspaceAutomator", headers=headers)

        # Raise an error if the API Key was not verified
        if message.status_code == 200:
            print("KEY WAS VERIFIED")
        else:
            raise API_KEY_ERROR(f"Failed verificaion [ERROR CODE: {message.status_code}]")

        # Encrypt API Key using password provided by user
        encrypted_key = Fernet(key_generator(self.key)).encrypt(new_key.encode()).decode()

        # Ask for confirmation
        if not inquirer.confirm("Want to save changes?") and not self.yes and self.command != "init":
            raise Exception("Exiting...")

        # Return encrypted API Key
        return encrypted_key

    ##### GitHub
    # Get GitHub username
    def get_github_username(self, **kwargs) -> str:

        while True:
            username = kwargs.get("github_user") or inquirer.text("Enter your GitHub username")

            # Ask for confirmation and return username
            if inquirer.confirm(f"Is {username} correct?") and self.__check_github_user(username=username):
                return username

    # Check if github user exists
    def __check_github_user(self, username: str) -> bool:

        if requests.get(f"https://github.com/{username}").status_code != 200:
            print("\nUsername does not exist\n")
            return False

        return True

    ##### SAVE SETTINGS
    # Save the settings
    @yaspin(text=" Saving settings...")
    @classmethod
    def save_settings(cls,
        settings: dict,
        **kwargs
        ) -> None:

        # Modify settings
        if settings:
            if kwargs.get("directories"):
                settings["directories"] = kwargs.get("directories")
            if kwargs.get("languages"):
                settings["languages"] = kwargs.get("languages")
            if kwargs.get("vscode"):
                settings["vscode"] = kwargs.get("vscode")
            if kwargs.get("github-user"):
                settings["github-user"] = kwargs.get("github-user")

        # Create settings
        elif kwargs.keys() == ["directories", "languages", "vscode", "github-user"]:
            settings = kwargs

        # Save and create files
        if settings:
            with open(SETTINGS, "w+") as f:
                json.dump(settings, f)

        if not os.path.exists(WORKSPACES):
            with open(WORKSPACES, "w+") as f:
                pass

        if kwargs.get("api-key"):
            with open(SECRETS, "w+") as f:
                f.write(f'API_KEY = "{kwargs.get("api-key").decode()}"')



########################################
#####  IMPORTING MODULES           #####
########################################

from pathlib import Path

########################################
#####  CODE                        #####
########################################

#####  GLOBAL VARIABLES
# Path to settings
settings_paths = {
    "Windows": str(Path.home()) + "\\AppData\\Roaming\\WorkSpaceAutomation",
    "Darwin": str(Path.home()) + "\\Library\\Application\\ Support\\WorkSpaceAutomation",
    "Linux": str(Path.home()) + "\\.config\\WorkSpaceAutomation"
}

# Path to vscode settings
code_paths = {
    "Windows": str(Path.home()) + "\\AppData\\Roaming\\{}\\User\\globalStorage\\storage.json",
    "Darwin": str(Path.home()) + "\\Library\\Application\\ Support\\{}\\User\\globalStorage\\storage.json",
    "Linux": str(Path.home()) + "\\.config\\{}\\User\\globalStorage\\storage.json"
}

# Languages
languages = {
    "Python": {
        "gitignore": "",
        "directories": [

        ],
        "files": [

        ],
        "venv-command": ""
    },
}
